Attribute VB_Name = "TraceFlowPaths"
Option Explicit

'=================================================================================
' Author: Carl Johnson (CJJ)
'
'----------------------------------
' Version 1.0 - Base version
' Date: 20/3/2008
' Description:
' creates and fills a PolylineZM feature class with lines originating from
' pourpoint locations and traces a line of maximum flow length from a raster.
'----------------------------------
'
'=================================================================================

' user input starts
' pour point layer name
Private Const PPOINTLAYERNAME As String = "SSP_00_snapped"
' field to use as an integer ID field in the pour point layer
Private Const CATCHIDFIELDNAME As String = "CatchID"

' flow length raster layer name
Private Const FLOWLENGTHLAYERNAME As String = "dem1m_wo_inlet4_bf_fl"
' watersheds raster layer name
Private Const CATCHMENTLAYERNAME As String = "dem1m_wo_inlet4_bf_ws1"
' DEM raster layer name
Private Const DEMLAYERNAME As String = "dem1m"  '"dem1m"

' user input ends

Private Dirs(0 To 7) As Integer
Private Const MSGBOXTITLE As String = "Trace Maximum Flow Length"

Sub TraceFlowPaths()
    
    ' new feature class variables
    Dim pGxDialog As IGxDialog
    Dim strCat As String, strName As String
    Dim pWSFactory As IWorkspaceFactory
    Dim pFWS As IFeatureWorkspace
    Dim pWS As IWorkspace
    Dim pFDataset As IFeatureDataset
    Dim pEnumDataset As IEnumDataset
    Dim pDataset As IDataset
    Dim pEnumDatasetName As IEnumDatasetName
    Dim pDatasetName As IDatasetName
    Dim IsFound As Boolean
    Dim pFCOut As IFeatureClass
    Dim pFieldsEdit As IFieldsEdit
    Dim pFieldEdit As IFieldEdit
    Dim pGeoDefEdit As IGeometryDefEdit
    Dim pGeoDef As IGeometryDef
    Dim pGeoDataset As IGeoDataset
    Dim lngCatField As Long, lngLengthField As Long, lngEASField As Long
    Dim pWSEdit As IWorkspaceEdit
        
    Dim pRCat As IRaster
    Dim pRFL As IRaster
    Dim pRDEM As IRaster
    Dim rpcCat As RasterPixelCacher
    Dim rpcFL As RasterPixelCacher
    Dim rpcDEM As RasterPixelCacher
    Dim pFC As IFeatureClass
    Dim pFLayer As IFeatureLayer
    
    Dim pZAware As IZAware
    Dim pMAware As IMAware
    
    Dim pEnumLayer As IEnumLayer
    Dim pLayer As ILayer
    Dim pMxDoc As IMxDocument
    Dim pRLayer As IRasterLayer
    Dim pRProps As IRasterProps
    Dim cont As Boolean
    Dim lngCatID As Long
    Dim intStepDir As Integer   ' 0 (right), 1 (bottom-right), 2 (bottom), 3 (bottom-left), 4 (left), etc.
    Dim i As Integer
    Dim pF As IFeature
    Dim pFOut As IFeature
    Dim pFCur As IFeatureCursor
    Dim pPoint As IPoint, pPoint2 As IPoint
    Dim pPL As IPolyline
    Dim pPointCol As IPointCollection
    Dim dX As Double, dY As Double
    Dim xMin As Double
    Dim xMax As Double
    Dim yMin As Double
    Dim yMax As Double
    
    ' equal area calculation variables
    Dim h As Double, hPrev As Double
    Dim x As Double, z0 As Double
    Dim xPrev As Double, deltaX As Double
    Dim h_bar As Double, A_accum As Double
    
    Set pMxDoc = ThisDocument
    Set pEnumLayer = pMxDoc.FocusMap.Layers(Nothing, True)
    pEnumLayer.Reset
    Set pLayer = pEnumLayer.Next
'    Set pRFA = Nothing
    Set pFC = Nothing
    While Not pLayer Is Nothing And (pRFL Is Nothing Or pFC Is Nothing Or pRCat Is Nothing Or pRDEM Is Nothing)
        If TypeOf pLayer Is IRasterLayer Then
            If pLayer.Name Like FLOWLENGTHLAYERNAME Then
                Set pRLayer = pLayer
                Set pRFL = pRLayer.Raster
            ElseIf pLayer.Name Like CATCHMENTLAYERNAME Then
                Set pRLayer = pLayer
                Set pRCat = pRLayer.Raster
            ElseIf pLayer.Name Like DEMLAYERNAME Then
                Set pRLayer = pLayer
                Set pRDEM = pRLayer.Raster
            End If
        ElseIf TypeOf pLayer Is IFeatureLayer Then
            If pLayer.Name Like PPOINTLAYERNAME Then
                Set pFLayer = pLayer
                If pFLayer.FeatureClass.ShapeType = esriGeometryPoint Then Set pFC = pFLayer.FeatureClass
            End If
        End If
        Set pLayer = pEnumLayer.Next
    Wend
    
    If pRFL Is Nothing Or pRCat Is Nothing Or pFC Is Nothing Or pRDEM Is Nothing Then
        MsgBox "Correct layers are not available. Macro terminating" & VBA.Chr(13) & "Maybe consider checking the layer names used in the VBA macro", vbYes, MSGBOXTITLE
        Exit Sub
    End If
    
    ' get the feature class to write extents to
    Set pGxDialog = New gxDialog
    With pGxDialog
        .AllowMultiSelect = False
        .ButtonCaption = "Save"
        .Name = "flow_paths_1"
        .Title = MSGBOXTITLE
        Set .ObjectFilter = New GxFilterFeatureClasses
    End With
            
    If pGxDialog.DoModalSave(0) Then
        ' get the workspace and the feature class to add features to
        strCat = pGxDialog.FinalLocation.Category
        strName = pGxDialog.Name
        If strCat Like "Folder" Then
            ' tidy up name
            If VBA.Right(strName, 4) Like ".shp" Then strName = VBA.Left(strName, Len(strName) - 4)
            ' shapefile workspace
            Set pWSFactory = New ShapefileWorkspaceFactory
            Set pFWS = pWSFactory.OpenFromFile(pGxDialog.FinalLocation.FullName, 0)
        ElseIf strCat Like "Personal Geodatabase" Then
            ' access database workspace
            Set pWSFactory = New AccessWorkspaceFactory
            Set pFWS = pWSFactory.OpenFromFile(pGxDialog.FinalLocation.FullName, 0)
        ElseIf strCat Like "Personal Geodatabase Feature Dataset" Then
            ' access database workspace featuredataset
            Set pWSFactory = New AccessWorkspaceFactory
            Set pFWS = pWSFactory.OpenFromFile(pGxDialog.FinalLocation.Parent.FullName, 0)
        Else
            MsgBox "The inputted featureclass's workspace is not supported"
            Exit Sub
        End If
    Else
        MsgBox "Macro terminating", vbOKOnly, MSGBOXTITLE
        Set pGxDialog = Nothing
        Exit Sub
    End If
        
    Set pWS = pFWS
           
    If strCat Like "Personal Geodatabase Feature Dataset" Then
        ' look for featureclass in featuredataset
        Set pFDataset = pFWS.OpenFeatureDataset(pGxDialog.FinalLocation.BaseName)
        IsFound = False
        Set pEnumDataset = pFDataset.Subsets
        pEnumDataset.Reset
        Set pDataset = pEnumDataset.Next
        While Not pDataset Is Nothing And Not IsFound
            IsFound = pDataset.Name Like strName
            Set pDataset = pEnumDataset.Next
        Wend
    Else
        ' look for featureclass in workspace
        IsFound = False
        Set pEnumDatasetName = pWS.DatasetNames(esriDTFeatureClass)
        pEnumDatasetName.Reset
        Set pDatasetName = pEnumDatasetName.Next
        While Not pDatasetName Is Nothing And Not IsFound
            IsFound = pDatasetName.Name Like strName
            Set pDatasetName = pEnumDatasetName.Next
        Wend
    End If

    If IsFound Then
        
        ' ask the user if they want to overwrite the existing feature class
        If MsgBox("Do you want to overwrite the existing featureclass?", vbYesNo, MSGBOXTITLE) = vbYes Then
            Set pDataset = pFWS.OpenFeatureClass(strName)
            On Error GoTo ErrorDeletingFC
            If pDataset.CanDelete Then
                pDataset.Delete
            Else
                GoTo DeletingFCOK
            End If
            GoTo DeletingFCOK
ErrorDeletingFC:
            MsgBox "Featureclass cannot be overwritten. Macro terminating", vbOKOnly, MSGBOXTITLE
            Exit Sub
DeletingFCOK:   On Error GoTo 0
        Else
            MsgBox "Macro terminating.", vbOKOnly, MSGBOXTITLE
            Exit Sub
        End If
    End If

    ' create a new feature class
    Set pFieldsEdit = New Fields
    
    ' OID field
    Set pFieldEdit = New Field
    With pFieldEdit
        .Editable = False
        .Name = "OBJECTID"
        .AliasName = "FID"
        .Type = esriFieldTypeOID
    End With
    pFieldsEdit.AddField pFieldEdit
    
    ' Shape field
    Set pFieldEdit = New Field
    Set pGeoDefEdit = New GeometryDef
    pGeoDefEdit.GeometryType = esriGeometryPolyline
    pGeoDefEdit.HasM = True
    pGeoDefEdit.HasZ = True
    
    If strCat Like "Personal Geodatabase Feature Dataset" Then
        ' spatial reference must be the same as the featuredataset
        Set pGeoDataset = pFDataset
        Set pGeoDefEdit.SpatialReference = pGeoDataset.SpatialReference
        Set pGeoDataset = Nothing
    Else
        ' set the spatial reference to be the same as the inputted featureclass
        Set pGeoDef = pFLayer.FeatureClass.Fields.Field(pFLayer.FeatureClass.FindField(pFLayer.FeatureClass.ShapeFieldName)).GeometryDef
        Set pGeoDefEdit.SpatialReference = pGeoDef.SpatialReference
        Set pGeoDef = Nothing
    End If
    
    With pFieldEdit
        .Editable = False
        .Name = "SHAPE"
        .Type = esriFieldTypeGeometry
        Set .GeometryDef = pGeoDefEdit
    End With
    pFieldsEdit.AddField pFieldEdit
    Set pGeoDefEdit = Nothing
    
    ' Featureclass name field
    Set pFieldEdit = New Field
    With pFieldEdit
        .Editable = True
        .Name = CATCHIDFIELDNAME
        .Type = esriFieldTypeInteger
    End With
    pFieldsEdit.AddField pFieldEdit
    
    ' length field
    Set pFieldEdit = New Field
    With pFieldEdit
        .Editable = True
        .Name = "H_Length"
        .Type = esriFieldTypeDouble
        .Precision = 16
        .Scale = 4
    End With
    pFieldsEdit.AddField pFieldEdit
    
    ' equal area slope field
    Set pFieldEdit = New Field
    With pFieldEdit
        .Editable = True
        .Name = "E_A_Slope"
        .Type = esriFieldTypeDouble
        .Precision = 16
        .Scale = 5
    End With
    pFieldsEdit.AddField pFieldEdit
    
    Set pFieldEdit = Nothing
    
    On Error GoTo FCCreationError
    
    ' create the new featureclass
    If strCat Like "Folder" Or strCat Like "Personal Geodatabase" Then
        Set pFCOut = pFWS.CreateFeatureClass(strName, pFieldsEdit, Nothing, Nothing, esriFTSimple, "SHAPE", "")
    ElseIf strCat Like "Personal Geodatabase Feature Dataset" Then
        Set pFCOut = pFDataset.CreateFeatureClass(strName, pFieldsEdit, Nothing, Nothing, esriFTSimple, "SHAPE", "")
    End If
    ' get index and length of the output field
    lngCatField = pFCOut.FindField(CATCHIDFIELDNAME)
    lngLengthField = pFCOut.FindField("H_Length")
    lngEASField = pFCOut.FindField("E_A_Slope")
    
    GoTo FCCreationOK
FCCreationError:
    MsgBox "Error occurred when creating a new featureclass. Macro terminating", vbOKOnly, MSGBOXTITLE
    Set pFieldsEdit = Nothing
    Set pFCOut = Nothing
    Exit Sub
FCCreationOK:
    On Error GoTo 0
             
    ' start adding features to featureclass
    Set pWSEdit = pWS
    Set pWS = Nothing
    pWSEdit.StartEditing False
   
    Set pGeoDataset = pFC
    xMin = pGeoDataset.Extent.xMin
    xMax = pGeoDataset.Extent.xMax
    yMin = pGeoDataset.Extent.yMin
    yMax = pGeoDataset.Extent.yMax
    Set pRProps = pRFL
    dX = pRProps.MeanCellSize.x
    dY = pRProps.MeanCellSize.y
    Set pRProps = Nothing
    
    ' give user feedback
    Application.StatusBar.ShowProgressBar "Trace Flow Paths from Pour Points", 0, pFC.FeatureCount(Nothing) - 1, 1, True
    Application.StatusBar.ProgressBar.Position = 0
    
    ' allow user to cancel process
    Dim pTrackCancel As ITrackCancel
    Set pTrackCancel = New CancelTracker
    pTrackCancel.CancelOnKeyPress = True
    pTrackCancel.CancelOnClick = False
        
    ' set up raster wrappers
    Set rpcCat = New RasterPixelCacher
    rpcCat.Raster = pRCat
    Set rpcFL = New RasterPixelCacher
    rpcFL.Raster = pRFL
    Set rpcDEM = New RasterPixelCacher
    rpcDEM.Raster = pRDEM
'    Set rpcFD = New RasterPixelCacher
'    rpcFD.Raster = pRFD
        
    Set pFCur = pFC.Search(Nothing, False)
    Set pF = pFCur.NextFeature
    While Not pF Is Nothing
        Set pPoint = pF.Shape
        lngCatID = pF.Value(pF.Fields.FindField(CATCHIDFIELDNAME))
        If lngCatID >= 0 Then
        
            Set pPL = New Polyline
            Set pZAware = pPL
            Set pMAware = pPL
            pZAware.ZAware = True
            pMAware.MAware = True
            Set pZAware = Nothing
            Set pMAware = Nothing
            
            ' trace up the current branch and pass back the geometry
            TracePath lngCatID, rpcFL, rpcCat, rpcDEM, pPoint.x, pPoint.y, pPL
            
            Application.StatusBar.StepProgressBar
            
            ' create a new output feature
            pWSEdit.StartEditOperation
            Set pFOut = pFCOut.CreateFeature
            Set pFOut.Shape = pPL
    
            ' add relevant info to the featureclass
            pFOut.Value(lngCatField) = lngCatID   'CLng(rpcCat.Value(0, pPoint.x, pPoint.y))
            pFOut.Value(lngLengthField) = pPL.Length
            
            ' calculate the equal area slope for this flow path
            ' assume the line runs from the bottom to the top
            z0 = pPL.FromPoint.Z
            xMax = pPL.FromPoint.M
            xPrev = 0
            hPrev = 0
            A_accum = 0
            Set pPointCol = pPL
            ' skip the first point
            For i = 1 To pPointCol.PointCount - 1
                Set pPoint2 = pPointCol.Point(i)
                h = pPoint2.Z - z0
                x = xMax - pPoint2.M
                deltaX = x - xPrev
                h_bar = (h + hPrev) / 2
                A_accum = A_accum + (h_bar * deltaX)
                ' save numbers for next iteration
                xPrev = x
                hPrev = h
            Next i
            If xMax < 0.001 Then
                pFOut.Value(lngEASField) = -1
            Else
                pFOut.Value(lngEASField) = 2 * A_accum / xMax / xMax
            End If
            
            Set pPL = Nothing
            
            pFOut.Store
            pWSEdit.StopEditOperation
            Set pFOut = Nothing
        
        End If
        
        Set pF = pFCur.NextFeature
        
        ' check cancel tracker
        If Not pTrackCancel.Continue Then
            Set pF = Nothing
            MsgBox "Macro has been cancelled by the user.", vbOKOnly, MSGBOXTITLE
        End If
    Wend
    
    Application.StatusBar.HideProgressBar
    
    pWSEdit.StopEditing True
    
    ' save the edits made
    Set pWSEdit = Nothing
    
    ' add layer to ArcMap
    Set pFLayer = New featureLayer
    Set pFLayer.FeatureClass = pFCOut
    pFLayer.Name = pFCOut.AliasName
    pMxDoc.FocusMap.AddLayer pFLayer
    
    Set pTrackCancel = Nothing
    Set pFLayer = Nothing
    Set pFC = Nothing
    Set pFCOut = Nothing
    
    Application.StatusBar.Message(0) = ""
    
    ' tell the user that the macro has completed
    MsgBox "Macro has returned successfully", vbOKOnly, MSGBOXTITLE
    
End Sub

Private Sub TracePath(lngCatNo As Long, rpcFL As RasterPixelCacher, rpcCat As RasterPixelCacher, rpcDEM As RasterPixelCacher, x As Double, y As Double, pPL As IPolyline)
    ' trace in a maximum-decreasing-value direction
    Dim dblFL As Double, dblFLTemp As Double
    Dim intDir As Integer
    Dim dblX As Double, dblY As Double
    Dim dblXNew As Double, dblYNew As Double
    Dim pRProps As IRasterProps
    Dim pPointCol As IPointCollection
    Dim pPoint As IPoint
    Dim cont As Boolean
    Dim dblNewMaxFL As Double
    Dim i As Integer
    Dim dX As Double, dY As Double
        
    Dim pZAware As IZAware
    Dim pMAware As IMAware
        
    ' centre x and y values
    dX = rpcFL.dX
    dY = rpcFL.dY
    Set pRProps = rpcFL.Raster
    dblX = pRProps.Extent.xMin + dX * (Fix((x - pRProps.Extent.xMin) / dX) + 0.5)
    dblY = pRProps.Extent.yMin + dY * (Fix((y - pRProps.Extent.yMin) / dY) + 0.5)
        
    ' add the first point to the polyline
    Set pPoint = New Point
    pPoint.PutCoords dblX, dblY
    
    ' get the catchment ID
    'lngCatNo = CLng(rpcCat.Value(0, dblX, dblY))
    ' get the current flow length
    dblFL = CDbl(rpcFL.Value(0, dblX, dblY))
    
    ' allow for cases where the sink points have been removed from the grid
    If dblFL < -3E+38 Then
        ' get a quick approximate start value as grid has a hole in it
        dblNewMaxFL = -1E+308
        dblFLTemp = -1E+308
        'lngCatNo = CLng(rpcCat.Value(0, dblX, dblY))
        intDir = -1
        cont = False
        For i = 0 To 7
            ' get the coordinates of the neighbouring cell
            StepCoords i, dX, dY, dblX, dblY, dblXNew, dblYNew
            ' check that the cell is in the same catchment
            If lngCatNo = rpcCat.Value(0, dblXNew, dblYNew) Then
                ' check that the flow length is less than the current value
                dblFLTemp = rpcFL.Value(0, dblXNew, dblYNew)
                ' test if it is the maximum so far
                If dblFLTemp > dblNewMaxFL Then
                    intDir = i
                    dblNewMaxFL = dblFLTemp
                    cont = True
                End If
            End If
        Next i
        If intDir Mod 2 = 0 Then
            StepCoords intDir, dX, dY, dblX, dblY, dblXNew, dblYNew
            pPoint.M = rpcFL.Value(0, dblXNew, dblYNew) + dX
        Else
            StepCoords intDir, dX, dY, dblX, dblY, dblXNew, dblYNew
            pPoint.M = rpcFL.Value(0, dblXNew, dblYNew) + Sqr((dX ^ 2) + (dY ^ 2))
        End If
        ' set the starting flow length
        dblFL = pPoint.M
    Else
        pPoint.M = rpcFL.Value(0, dblX, dblY)
        cont = True
    End If
    
    ' set the point elevation and add this to the collection
    pPoint.Z = rpcDEM.Value(0, dblX, dblY)
    Set pPointCol = pPL
    pPointCol.AddPoint pPoint
    Set pPoint = Nothing
        
    ' find the direction of the next highest value
    ' if there are identical flow lengths then use the first one tested
    While cont
        ' look for the next highest flow length value in the surrounding cells
        dblNewMaxFL = -1E+308
        intDir = -1
        cont = False
        For i = 0 To 7
            ' get the coordinates of the neighbouring cell
            StepCoords i, dX, dY, dblX, dblY, dblXNew, dblYNew
            ' check that the cell is in the same catchment
            If lngCatNo = rpcCat.Value(0, dblXNew, dblYNew) Then
                ' check that the flow length is less than the current value
                dblFLTemp = rpcFL.Value(0, dblXNew, dblYNew)
                If dblFLTemp < dblFL Then
                    ' test if it is the maximum so far
                    If dblFLTemp > dblNewMaxFL Then
                        intDir = i
                        dblNewMaxFL = dblFLTemp
                    End If
                    ' continue the trace as a maximum has been found
                    cont = True
                End If
            End If
        Next i
        
        If cont Then
            ' a new step has been made so add it to the polyline
            Set pPoint = New Point
            Set pZAware = pPoint
            Set pMAware = pPoint
            pZAware.ZAware = True
            pMAware.MAware = True
            Set pZAware = Nothing
            Set pMAware = Nothing
                        
            StepCoords intDir, dX, dY, dblX, dblY, dblXNew, dblYNew
            pPoint.PutCoords dblXNew, dblYNew
            pPoint.M = rpcFL.Value(0, dblXNew, dblYNew)
            pPoint.Z = rpcDEM.Value(0, dblXNew, dblYNew)
            pPointCol.AddPoint pPoint
            Set pPoint = Nothing
            
            dblX = dblXNew
            dblY = dblYNew
            dblFL = rpcFL.Value(0, dblX, dblY)
        End If
    Wend

End Sub

Sub StepCoords(dir As Integer, dX As Double, dY As Double, xIn As Double, yIn As Double, ByRef xOut As Double, ByRef yOut As Double)
    Select Case dir
        Case 0
            xOut = xIn + dX
            yOut = yIn
        Case 1
            xOut = xIn + dX
            yOut = yIn - dY
        Case 2
            xOut = xIn
            yOut = yIn - dY
        Case 3
            xOut = xIn - dX
            yOut = yIn - dY
        Case 4
            xOut = xIn - dX
            yOut = yIn
        Case 5
            xOut = xIn - dX
            yOut = yIn + dY
        Case 6
            xOut = xIn
            yOut = yIn + dY
        Case 7
            xOut = xIn + dX
            yOut = yIn + dY
        Case Else
            MsgBox "Unknown step direction: Value = " & dir, vbOKOnly, MSGBOXTITLE
            xOut = xIn
            yOut = yIn
    End Select
End Sub


